/***************************************************************************
 Prctica n 11	   Simulador de redes neuronales formales.

 Fecha: 21 Enero 1997

 Autor: Oscar Martnez Valpuesta

 Objetivo:   Se pretende realizar un simulador de redes neuronales que
	   utilice el modelo de neurona de McCulloch y Pitts.

 Comentarios:   La red neuronal se va a almacenar en un vector dinmico (de
	      tamao n de neuronas + n de entradas) en el que cada celda
	      va a ser un registro con la informacin del estado actual,
	      el estado al que va a transitar, el umbral y dos listas: una
	      con las entradas excitatorias y otra con las entradas
	      inhibitorias. Las entradas exteriores las vamos a considerar
	      como neuronas para poder introducirlas dentro de la estructura.

		Modelo de neurona de McCulloch y Pitts. Cada neurona tiene una
	      sola salida, que una vez abandonado el cuerpo se pude ramificar
	      y servir como entrada a otras neuronas o a ella misma. Cada
	      neurona puede tener cualquier nmero de terminales de entrada
	      que pueden ser: entradas excitatorias o entradas inhibitorias.
	      Una neurona puede cambiar su estado interno entre encendido y
	      apagado. Esto depende del nmero de entradas excitatorias activas
	      (umbral) y si existe alguna entrada inhibitoria activa (la
	      neurona estara desactivada).

		El simulador carga desde un fichero la estructura de la red
	      a simular y desde otro las entradas que tendr dicha red a lo
	      largo del tiempo.

 Salida:   El programa visualiza en pantalla un tabla con el valor de las
	 entradas y estado de las neuronas a medida que evoluciona la red.
****************************************************************************/
#include "util.h"	/* incluye stdio.h (printf, scanf, fscanf) y conio.h */
#include <stdlib.h>	/* malloc, calloc */

/********************************* TIPOS ***********************************/
enum conj_est {INACTIVA = 0, ACTIVA = 1};	/* estados de las neuronas */

typedef struct nodo_d {	  /* representacin de una dendrita */
  unsigned neu;		  /* neurona/entrada de la que viene la dendrita */
  struct nodo_d *sig;     /* siguiente dendrita */
} dendrita;

typedef struct {	     /* representacin de una neurona */
  enum conj_est estado[2];   /* estados actual y nuevo de la neurona */
  unsigned short umbral;     /* umbral de la neurona */
  dendrita *inh, *exi;       /* lista de entradas inhib. y excit. de la neurona */
} neurona;

/****************************** VARIABLES **********************************/
unsigned total_ne;	/* n total de neuronas + entradas */
neurona *red;           /* red neuronal */
unsigned instante;	/* instantes de tiempo */
unsigned entradas = 0;	/* n de entradas a la red */

/****************************** PROTOTIPOS *********************************/
void err_formato(void);
void lee_red(char *nom);
void lee_entradas(FILE *fd);
void cambiar_red(void);
void imprimir_cabecera(void);
void imprimir_red(void);
void itera_red(char *nom);

/***************************************************************************
   Funcin que visualiza en pantalla que se ha producido un error en el
 formato del fichero de definicin de la red.
****************************************************************************/
void err_formato(void) {

  printf("\n\nERROR, se ha producido un error en la lectura del fichero de definicin de la red.");
  exit(0);
}

/***************************************************************************
   Lee el fichero con las caractersticas de la red y lo carga en memoria
 dinmica, inicializando sus estados de activacin a nulo.
   Las entradas exteriores las vamos a considerar como neuronas y as poder
 incluirlas dentro de la estructura.
   Las entradas a una neurona tiene que estar entre [-total_ne, total_ne], si
 no se produce un error.
****************************************************************************/
void lee_red(char *nom) {
FILE *fichero;          /* handler para el fichero de definicin de la red */
int aux;                /* variable para la lectura de diversos datos */
dendrita *nodo;         /* entrada a una neurona */
register unsigned i;
boolean primera;        /* indica si es la primera neurona que se lee */
char ch;		/* controla que se llega al final de lnea en fichero */

  /* apertura del fichero de definicin de la red */
  fichero = fopen(nom, "r");
  if (fichero == NULL) {
    ERROR1(nom);
    exit(0);
  }

  fscanf(fichero, "%ud", &total_ne);	/* lectura de neuronas+entradas */

  /* lectura del n de entradas */
  fscanf(fichero, "%d", &aux);
  while (aux == -1) {
    entradas++;
    fscanf(fichero, "%d", &aux);
  }
  /* memoria para la red */
  red = (neurona *)calloc(total_ne, sizeof(neurona));
  if (red == NULL) {
    ERROR_MEM;
    exit(0);
  }

  /* lectura de las neuronas */
  primera = TRUE;
  for (i = entradas; i < total_ne; i++) {
    if (!primera)
      fscanf(fichero, "%d", &aux);  /* lectura del umbral */
    primera = FALSE;
    red[i].umbral = (unsigned short)aux;
    ch = '0';
/* mientras no se llegue al final de lnea se leen las entradas a la neurona */
    while (ch != '\n') {
      fscanf(fichero, "%d", &aux);
      if ((aux <= (int)(total_ne)) && (aux >= ((int)(-1 * total_ne)))) {
	/* memoria para la entrada a la neurona */
	nodo = (dendrita *)malloc(sizeof(dendrita));
	if (nodo == NULL) {
	  ERROR_MEM;
	  exit(0);
	}

	if (aux > 0) {	/* entrada excitatoria */
	  nodo -> neu = (unsigned) aux;
	  nodo -> sig = red[i].exi;
	  red[i].exi = nodo;
	}
	else
	  if (aux < 0) { /* entrada inhibitoria */
	    nodo -> neu = (unsigned) (-1 * aux);
	    nodo -> sig = red[i].inh;
	    red[i].inh = nodo;
	  }
      }
      else {
	err_formato();
	exit(0);
      }
      fscanf(fichero,"%c", &ch);
    } /* end while */
  } /* end for */
}

/***************************************************************************
   Funcin que lee una fila de entradas desde el fichero de entradas y las
 carga en el estado de las neuronas correspondientes a entradas. El valor
 de la entrada se almacena en el estado actual de la neurona ficticia, esto
 es, en estado[0].
****************************************************************************/
void lee_entradas(FILE *fd) {
register unsigned i;


  for (i = 0; i < entradas; i++)
    fscanf(fd, "%d", &red[i].estado[0]);
}

/***************************************************************************
   Cambia los estados de todas las neuronas de la red segn las reglas
 descritas por McCulloch y Pitts:

 si hay alguna entrada inhibitoria activa
   neurona inactiva
 si el n de entradas excitatorias activas es >= umbral
   neurona activa
 sino
   neurona inactiva
****************************************************************************/
void cambiar_red(void) {
register unsigned i;
unsigned short on;   /* n de dendritas excitatorias activadas */
boolean inhib;	     /* indica si hay alguna dendrita inhibitoria activada */
dendrita *nodo;      /* recorre las listas de entradas exi. e inh. */

  for (i = entradas; i < total_ne; i++) {
    inhib = FALSE;
    nodo = red[i].inh;
    red[i].estado[1] = INACTIVA; /* por defecto, el nuevo estado es inactiva */
    /* se busca se hay alguna entrada inhibitoria activada */
    while ((nodo != NULL) && (!inhib))
      if (red[nodo -> neu - 1].estado[0])
	inhib = TRUE;
      else
	nodo = nodo -> sig;

    if (!inhib) {	/* se comprueba el n de entradas excitatorias activas */
      nodo = red[i].exi;
      on = 0;
      while ((nodo != NULL) && (on < red[i].umbral)) {
	if (red[nodo -> neu - 1].estado[0])
	  on++;
	nodo = nodo -> sig;
      }
      /* para activar la neurona tinene que haber como mnimo el umbral */
      if (on == red[i].umbral)
	red[i].estado[1] = ACTIVA;
    }
  }

  instante++;
  /* se pasan los estados nuevos a los estados actuales */
  for (i = entradas; i < total_ne; i++)
    red[i].estado[0] = red[i].estado[1];
}

/***************************************************************************
   Imprime la cabecera de la tabla de estados de las neuronas.
****************************************************************************/
void imprimir_cabecera(void) {
register unsigned i;

  printf("\n   Entradas: %d\t\tNeuronas: %d", entradas, total_ne - entradas);
  printf("\n t");
  for (i = 1; i <= total_ne; i++)
    printf("    %d", i);
  printf("\n");
  for (i = 0; i <= 5 * total_ne; i++)
    printf("");
  /* tiene en cuenta nmeros de dos cifras */
  if (total_ne >= 9)
    for (i = 0; i <= total_ne - 10; i++)
      printf("");
}

/***************************************************************************
   Imprime la tabla de estados de las neuronas para el instante actual.
****************************************************************************/
void imprimir_red(void) {
register unsigned i;

  printf("\n%02d", instante);
  for (i = 0; i < total_ne; i++)
    if ( i < 9)
      printf("    %d", red[i].estado[0]);
    else   /* tiene en cuemta nmeros de dos cifras */
      printf("     %d", red[i].estado[0]);
}

/***************************************************************************
   Funcin que se encarga de coordinar el funcionamiento de la red.
****************************************************************************/
void itera_red(char *nom) {
FILE *fichero;		/* handler para fichero de entradas */

  /* apertura del fichero de entradas */
  fichero = fopen(nom, "r");
  if (fichero == NULL) {
    ERROR1(nom);
    exit(0);
  }

  imprimir_cabecera();
  instante = 0;
  while (!feof(fichero)) {
    lee_entradas(fichero);
    imprimir_red();
    cambiar_red();
  }
  Pausa();
}


/***************************************************************************/
int main(int argc, char *argv[]) {

  clrscr();
  if (argc != 3){
	 printf("\nERROR, forma de uso: <definicin.RED> <entradas.IN>\n");
	 printf("\ndefinicin.RED: fichero de definicin de la red.");
	 printf("\nentradas.IN: fichero con las entras a la red.");
	 exit(0);
  }

  lee_red(argv[1]);
  itera_red(argv[2]);

  return 0;
}
